@@ -1,17 +1,17 @@
+use crate::prelude::*;
+use crate::runtime::vm::{RuntimeLinearMemory, VMMemoryImport};
 use crate::store::{StoreData, StoreOpaque, Stored};
 use crate::trampoline::generate_memory_export;
 use crate::Trap;
 use crate::{AsContext, AsContextMut, Engine, MemoryType, StoreContext, StoreContextMut};
-use anyhow::{bail, Result};
-use std::cell::UnsafeCell;
-use std::convert::TryFrom;
-use std::ops::Range;
-use std::slice;
-use std::time::Instant;
+use core::cell::UnsafeCell;
+use core::fmt;
+use core::ops::Range;
+use core::slice;
+use core::time::Duration;
 use wasmtime_environ::MemoryPlan;
-use wasmtime_runtime::{RuntimeLinearMemory, VMMemoryImport};
 
-pub use wasmtime_runtime::WaitResult;
+pub use crate::runtime::vm::WaitResult;
 
 /// Error for out of bounds [`Memory`] access.
 #[derive(Debug)]
@@ -21,12 +21,13 @@ pub struct MemoryAccessError {
     _private: (),
 }
 
-impl std::fmt::Display for MemoryAccessError {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+impl fmt::Display for MemoryAccessError {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         write!(f, "out of bounds memory access")
     }
 }
 
+#[cfg(feature = "std")]
 impl std::error::Error for MemoryAccessError {}
 
 /// A WebAssembly linear memory.
@@ -202,7 +203,7 @@ impl std::error::Error for MemoryAccessError {}
 /// be provided.
 #[derive(Copy, Clone, Debug)]
 #[repr(transparent)] // here for the C API
-pub struct Memory(Stored<wasmtime_runtime::ExportMemory>);
+pub struct Memory(Stored<crate::runtime::vm::ExportMemory>);
 
 impl Memory {
     /// Creates a new WebAssembly memory given the configuration of `ty`.
@@ -238,7 +239,6 @@ impl Memory {
         Self::_new(store.as_context_mut().0, ty)
     }
 
-    #[cfg_attr(nightlydoc, doc(cfg(feature = "async")))]
     /// Async variant of [`Memory::new`]. You must use this variant with
     /// [`Store`](`crate::Store`)s which have a
     /// [`ResourceLimiterAsync`](`crate::ResourceLimiterAsync`).
@@ -430,7 +430,17 @@ impl Memory {
 
     /// Returns the byte length of this memory.
     ///
-    /// The returned value will be a multiple of the wasm page size, 64k.
+    /// WebAssembly memories are made up of a whole number of pages, so the byte
+    /// size returned will always be a multiple of this memory's page size. Note
+    /// that different Wasm memories may have different page sizes. You can get
+    /// a memory's page size via the [`Memory::page_size`] method.
+    ///
+    /// By default the page size is 64KiB (aka `0x10000`, `2**16`, `1<<16`, or
+    /// `65536`) but [the custom-page-sizes proposal] allows a memory to opt
+    /// into a page size of `1`. Future extensions might allow any power of two
+    /// as a page size.
+    ///
+    /// [the custom-page-sizes proposal]: https://github.com/WebAssembly/custom-page-sizes
     ///
     /// For more information and examples see the documentation on the
     /// [`Memory`] type.
@@ -446,7 +456,19 @@ impl Memory {
         unsafe { (*store[self.0].definition).current_length() }
     }
 
-    /// Returns the size, in WebAssembly pages, of this wasm memory.
+    /// Returns the size, in units of pages, of this Wasm memory.
+    ///
+    /// WebAssembly memories are made up of a whole number of pages, so the byte
+    /// size returned will always be a multiple of this memory's page size. Note
+    /// that different Wasm memories may have different page sizes. You can get
+    /// a memory's page size via the [`Memory::page_size`] method.
+    ///
+    /// By default the page size is 64KiB (aka `0x10000`, `2**16`, `1<<16`, or
+    /// `65536`) but [the custom-page-sizes proposal] allows a memory to opt
+    /// into a page size of `1`. Future extensions might allow any power of two
+    /// as a page size.
+    ///
+    /// [the custom-page-sizes proposal]: https://github.com/WebAssembly/custom-page-sizes
     ///
     /// # Panics
     ///
@@ -456,7 +478,48 @@ impl Memory {
     }
 
     pub(crate) fn internal_size(&self, store: &StoreOpaque) -> u64 {
-        (self.internal_data_size(store) / wasmtime_environ::WASM_PAGE_SIZE as usize) as u64
+        let byte_size = self.internal_data_size(store);
+        let page_size = usize::try_from(self._page_size(store)).unwrap();
+        u64::try_from(byte_size / page_size).unwrap()
+    }
+
+    /// Returns the size of a page, in bytes, for this memory.
+    ///
+    /// WebAssembly memories are made up of a whole number of pages, so the byte
+    /// size (as returned by [`Memory::data_size`]) will always be a multiple of
+    /// their page size. Different Wasm memories may have different page sizes.
+    ///
+    /// By default this is 64KiB (aka `0x10000`, `2**16`, `1<<16`, or `65536`)
+    /// but [the custom-page-sizes proposal] allows opting into a page size of
+    /// `1`. Future extensions might allow any power of two as a page size.
+    ///
+    /// [the custom-page-sizes proposal]: https://github.com/WebAssembly/custom-page-sizes
+    pub fn page_size(&self, store: impl AsContext) -> u64 {
+        self._page_size(store.as_context().0)
+    }
+
+    pub(crate) fn _page_size(&self, store: &StoreOpaque) -> u64 {
+        store[self.0].memory.memory.page_size()
+    }
+
+    /// Returns the log2 of this memory's page size, in bytes.
+    ///
+    /// WebAssembly memories are made up of a whole number of pages, so the byte
+    /// size (as returned by [`Memory::data_size`]) will always be a multiple of
+    /// their page size. Different Wasm memories may have different page sizes.
+    ///
+    /// By default the page size is 64KiB (aka `0x10000`, `2**16`, `1<<16`, or
+    /// `65536`) but [the custom-page-sizes proposal] allows opting into a page
+    /// size of `1`. Future extensions might allow any power of two as a page
+    /// size.
+    ///
+    /// [the custom-page-sizes proposal]: https://github.com/WebAssembly/custom-page-sizes
+    pub fn page_size_log2(&self, store: impl AsContext) -> u8 {
+        self._page_size_log2(store.as_context().0)
+    }
+
+    pub(crate) fn _page_size_log2(&self, store: &StoreOpaque) -> u8 {
+        store[self.0].memory.memory.page_size_log2
     }
 
     /// Grows this WebAssembly memory by `delta` pages.
@@ -469,6 +532,13 @@ impl Memory {
     /// On success returns the number of pages this memory previously had
     /// before the growth succeeded.
     ///
+    /// Note that, by default, a WebAssembly memory's page size is 64KiB (aka
+    /// 65536 or 2<sup>16</sup>). The [custom-page-sizes proposal] allows Wasm
+    /// memories to opt into a page size of one byte (and this may be further
+    /// relaxed to any power of two in a future extension).
+    ///
+    /// [custom-page-sizes proposal]: https://github.com/WebAssembly/custom-page-sizes
+    ///
     /// # Errors
     ///
     /// Returns an error if memory could not be grown, for example if it exceeds
@@ -513,14 +583,14 @@ impl Memory {
                 Some(size) => {
                     let vm = (*mem).vmmemory();
                     *store[self.0].definition = vm;
-                    Ok(u64::try_from(size).unwrap() / u64::from(wasmtime_environ::WASM_PAGE_SIZE))
+                    let page_size = (*mem).page_size();
+                    Ok(u64::try_from(size).unwrap() / page_size)
                 }
                 None => bail!("failed to grow memory by `{}`", delta),
             }
         }
     }
 
-    #[cfg_attr(nightlydoc, doc(cfg(feature = "async")))]
     /// Async variant of [`Memory::grow`]. Required when using a
     /// [`ResourceLimiterAsync`](`crate::ResourceLimiterAsync`).
     ///
@@ -545,17 +615,17 @@ impl Memory {
         store.on_fiber(|store| self.grow(store, delta)).await?
     }
 
-    fn wasmtime_memory(&self, store: &mut StoreOpaque) -> *mut wasmtime_runtime::Memory {
+    fn wasmtime_memory(&self, store: &mut StoreOpaque) -> *mut crate::runtime::vm::Memory {
         unsafe {
             let export = &store[self.0];
-            wasmtime_runtime::Instance::from_vmctx(export.vmctx, |handle| {
+            crate::runtime::vm::Instance::from_vmctx(export.vmctx, |handle| {
                 handle.get_defined_memory(export.index)
             })
         }
     }
 
     pub(crate) unsafe fn from_wasmtime_memory(
-        wasmtime_export: wasmtime_runtime::ExportMemory,
+        wasmtime_export: crate::runtime::vm::ExportMemory,
         store: &mut StoreOpaque,
     ) -> Memory {
         Memory(store.store_data_mut().insert(wasmtime_export))
@@ -565,9 +635,9 @@ impl Memory {
         &store[self.0].memory.memory
     }
 
-    pub(crate) fn vmimport(&self, store: &StoreOpaque) -> wasmtime_runtime::VMMemoryImport {
+    pub(crate) fn vmimport(&self, store: &StoreOpaque) -> crate::runtime::vm::VMMemoryImport {
         let export = &store[self.0];
-        wasmtime_runtime::VMMemoryImport {
+        crate::runtime::vm::VMMemoryImport {
             from: export.definition,
             vmctx: export.vmctx,
             index: export.index,
@@ -583,7 +653,7 @@ impl Memory {
     /// Even if the same underlying memory definition is added to the
     /// `StoreData` multiple times and becomes multiple `wasmtime::Memory`s,
     /// this hash key will be consistent across all of these memories.
-    pub(crate) fn hash_key(&self, store: &StoreOpaque) -> impl std::hash::Hash + Eq {
+    pub(crate) fn hash_key(&self, store: &StoreOpaque) -> impl core::hash::Hash + Eq {
         store[self.0].definition as usize
     }
 }
@@ -720,12 +790,17 @@ pub unsafe trait MemoryCreator: Send + Sync {
 /// # }
 /// ```
 #[derive(Clone)]
-pub struct SharedMemory(wasmtime_runtime::SharedMemory, Engine);
+pub struct SharedMemory {
+    vm: crate::runtime::vm::SharedMemory,
+    engine: Engine,
+    page_size_log2: u8,
+}
 
 impl SharedMemory {
     /// Construct a [`SharedMemory`] by providing both the `minimum` and
     /// `maximum` number of 64K-sized pages. This call allocates the necessary
     /// pages on the system.
+    #[cfg(feature = "threads")]
     pub fn new(engine: &Engine, ty: MemoryType) -> Result<Self> {
         if !ty.is_shared() {
             bail!("shared memory must have the `shared` flag enabled on its memory type")
@@ -733,19 +808,39 @@ impl SharedMemory {
         debug_assert!(ty.maximum().is_some());
 
         let tunables = engine.tunables();
-        let plan = MemoryPlan::for_memory(ty.wasmtime_memory().clone(), tunables);
-        let memory = wasmtime_runtime::SharedMemory::new(plan)?;
-        Ok(Self(memory, engine.clone()))
+        let plan = MemoryPlan::for_memory(*ty.wasmtime_memory(), tunables);
+        let page_size_log2 = plan.memory.page_size_log2;
+        let memory = crate::runtime::vm::SharedMemory::new(plan)?;
+
+        Ok(Self {
+            vm: memory,
+            engine: engine.clone(),
+            page_size_log2,
+        })
     }
 
     /// Return the type of the shared memory.
     pub fn ty(&self) -> MemoryType {
-        MemoryType::from_wasmtime_memory(&self.0.ty())
+        MemoryType::from_wasmtime_memory(&self.vm.ty())
     }
 
     /// Returns the size, in WebAssembly pages, of this wasm memory.
     pub fn size(&self) -> u64 {
-        (self.data_size() / wasmtime_environ::WASM_PAGE_SIZE as usize) as u64
+        let byte_size = u64::try_from(self.data_size()).unwrap();
+        let page_size = u64::from(self.page_size());
+        byte_size / page_size
+    }
+
+    /// Returns the size of a page, in bytes, for this memory.
+    ///
+    /// By default this is 64KiB (aka `0x10000`, `2**16`, `1<<16`, or `65536`)
+    /// but [the custom-page-sizes proposal] allows opting into a page size of
+    /// `1`. Future extensions might allow any power of two as a page size.
+    ///
+    /// [the custom-page-sizes proposal]: https://github.com/WebAssembly/custom-page-sizes
+    pub fn page_size(&self) -> u32 {
+        debug_assert!(self.page_size_log2 == 0 || self.page_size_log2 == 16);
+        1 << self.page_size_log2
     }
 
     /// Returns the byte length of this memory.
@@ -755,7 +850,7 @@ impl SharedMemory {
     /// For more information and examples see the documentation on the
     /// [`Memory`] type.
     pub fn data_size(&self) -> usize {
-        self.0.byte_size()
+        self.vm.byte_size()
     }
 
     /// Return access to the available portion of the shared memory.
@@ -780,7 +875,7 @@ impl SharedMemory {
     /// currently be done unsafely.
     pub fn data(&self) -> &[UnsafeCell<u8>] {
         unsafe {
-            let definition = &*self.0.vmmemory_ptr();
+            let definition = &*self.vm.vmmemory_ptr();
             slice::from_raw_parts(definition.base.cast(), definition.current_length())
         }
     }
@@ -802,11 +897,11 @@ impl SharedMemory {
     /// [`ResourceLimiter`](crate::ResourceLimiter) is another example of
     /// preventing a memory to grow.
     pub fn grow(&self, delta: u64) -> Result<u64> {
-        match self.0.grow(delta, None)? {
+        match self.vm.grow(delta, None)? {
             Some((old_size, _new_size)) => {
                 // For shared memory, the `VMMemoryDefinition` is updated inside
                 // the locked region.
-                Ok(u64::try_from(old_size).unwrap() / u64::from(wasmtime_environ::WASM_PAGE_SIZE))
+                Ok(u64::try_from(old_size).unwrap() / u64::from(self.page_size()))
             }
             None => bail!("failed to grow memory by `{}`", delta),
         }
@@ -829,7 +924,7 @@ impl SharedMemory {
     /// This function will return an error if `addr` is not within bounds or
     /// not aligned to a 4-byte boundary.
     pub fn atomic_notify(&self, addr: u64, count: u32) -> Result<u32, Trap> {
-        self.0.atomic_notify(addr, count)
+        self.vm.atomic_notify(addr, count)
     }
 
     /// Equivalent of the WebAssembly `memory.atomic.wait32` instruction for
@@ -844,9 +939,8 @@ impl SharedMemory {
     /// the byte address `addr` specified. The `addr` specified is an index
     /// into this linear memory.
     ///
-    /// The optional `timeout` argument is the point in time after which the
-    /// calling thread is guaranteed to be woken up. Blocking will not occur
-    /// past this point.
+    /// The optional `timeout` argument is the maximum amount of time to block
+    /// the current thread. If not specified the thread may sleep indefinitely.
     ///
     /// This function returns one of three possible values:
     ///
@@ -870,9 +964,9 @@ impl SharedMemory {
         &self,
         addr: u64,
         expected: u32,
-        timeout: Option<Instant>,
+        timeout: Option<Duration>,
     ) -> Result<WaitResult, Trap> {
-        self.0.atomic_wait32(addr, expected, timeout)
+        self.vm.atomic_wait32(addr, expected, timeout)
     }
 
     /// Equivalent of the WebAssembly `memory.atomic.wait64` instruction for
@@ -888,21 +982,21 @@ impl SharedMemory {
         &self,
         addr: u64,
         expected: u64,
-        timeout: Option<Instant>,
+        timeout: Option<Duration>,
     ) -> Result<WaitResult, Trap> {
-        self.0.atomic_wait64(addr, expected, timeout)
+        self.vm.atomic_wait64(addr, expected, timeout)
     }
 
     /// Return a reference to the [`Engine`] used to configure the shared
     /// memory.
     pub(crate) fn engine(&self) -> &Engine {
-        &self.1
+        &self.engine
     }
 
     /// Construct a single-memory instance to provide a way to import
     /// [`SharedMemory`] into other modules.
-    pub(crate) fn vmimport(&self, store: &mut StoreOpaque) -> wasmtime_runtime::VMMemoryImport {
-        let export_memory = generate_memory_export(store, &self.ty(), Some(&self.0)).unwrap();
+    pub(crate) fn vmimport(&self, store: &mut StoreOpaque) -> crate::runtime::vm::VMMemoryImport {
+        let export_memory = generate_memory_export(store, &self.ty(), Some(&self.vm)).unwrap();
         VMMemoryImport {
             from: export_memory.definition,
             vmctx: export_memory.vmctx,
@@ -914,25 +1008,34 @@ impl SharedMemory {
     /// function is available to handle the case in which a Wasm module exports
     /// shared memory and the user wants host-side access to it.
     pub(crate) unsafe fn from_wasmtime_memory(
-        wasmtime_export: wasmtime_runtime::ExportMemory,
+        wasmtime_export: crate::runtime::vm::ExportMemory,
         store: &mut StoreOpaque,
     ) -> Self {
-        wasmtime_runtime::Instance::from_vmctx(wasmtime_export.vmctx, |handle| {
+        #[cfg_attr(not(feature = "threads"), allow(unused_variables, unreachable_code))]
+        crate::runtime::vm::Instance::from_vmctx(wasmtime_export.vmctx, |handle| {
+            let memory_index = handle.module().memory_index(wasmtime_export.index);
+            let page_size = handle.memory_page_size(memory_index);
+            debug_assert!(page_size.is_power_of_two());
+            let page_size_log2 = u8::try_from(page_size.ilog2()).unwrap();
+
             let memory = handle
                 .get_defined_memory(wasmtime_export.index)
                 .as_mut()
                 .unwrap();
-            let shared_memory = memory
-                .as_shared_memory()
-                .expect("unable to convert from a shared memory")
-                .clone();
-            Self(shared_memory, store.engine().clone())
+            match memory.as_shared_memory() {
+                Some(mem) => Self {
+                    vm: mem.clone(),
+                    engine: store.engine().clone(),
+                    page_size_log2,
+                },
+                None => panic!("unable to convert from a shared memory"),
+            }
         })
     }
 }
 
-impl std::fmt::Debug for SharedMemory {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+impl fmt::Debug for SharedMemory {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("SharedMemory").finish_non_exhaustive()
     }
 }
