@@ -1,3 +1,6 @@
+use crate::prelude::*;
+use crate::runtime::vm::{CompiledModuleId, MmapVec, ModuleMemoryImages, VMWasmCallFunction};
+use crate::sync::OnceLock;
 use crate::{
     code::CodeObject,
     code_memory::CodeMemory,
@@ -7,28 +10,21 @@ use crate::{
     types::{ExportType, ExternType, ImportType},
     Engine,
 };
-use anyhow::{bail, Context, Result};
-use once_cell::sync::OnceCell;
-use std::fs;
-use std::mem;
-use std::ops::Range;
+use alloc::sync::Arc;
+use core::fmt;
+use core::ops::Range;
+use core::ptr::NonNull;
+#[cfg(feature = "std")]
 use std::path::Path;
-use std::ptr::NonNull;
-use std::sync::Arc;
 use wasmparser::{Parser, ValidPayload, Validator};
 use wasmtime_environ::{
-    CompiledModuleInfo, DefinedFuncIndex, DefinedMemoryIndex, EntityIndex, HostPtr, ModuleTypes,
-    ObjectKind, VMOffsets,
+    CompiledModuleInfo, EntityIndex, HostPtr, ModuleTypes, ObjectKind, TypeTrace, VMOffsets,
+    VMSharedTypeIndex,
 };
-use wasmtime_runtime::{
-    CompiledModuleId, MemoryImage, MmapVec, ModuleMemoryImages, VMArrayCallFunction,
-    VMNativeCallFunction, VMSharedTypeIndex, VMWasmCallFunction,
-};
-
 mod registry;
 
 pub use registry::{
-    is_wasm_trap_pc, register_code, unregister_code, ModuleRegistry, RegisteredModuleId,
+    lookup_code, register_code, unregister_code, ModuleRegistry, RegisteredModuleId,
 };
 
 /// A compiled WebAssembly module, ready to be instantiated.
@@ -40,14 +36,25 @@ pub use registry::{
 /// representation. Instead you'll need to create an
 /// [`Instance`](crate::Instance) to interact with the wasm module.
 ///
-/// Creating a `Module` currently involves compiling code, meaning that it can
-/// be an expensive operation. All `Module` instances are compiled according to
-/// the configuration in [`Config`], but typically they're JIT-compiled. If
-/// you'd like to instantiate a module multiple times you can do so with
-/// compiling the original wasm module only once with a single [`Module`]
-/// instance.
+/// A `Module` can be created by compiling WebAssembly code through APIs such as
+/// [`Module::new`]. This would be a JIT-style use case where code is compiled
+/// just before it's used. Alternatively a `Module` can be compiled in one
+/// process and [`Module::serialize`] can be used to save it to storage. A later
+/// call to [`Module::deserialize`] will quickly load the module to execute and
+/// does not need to compile any code, representing a more AOT-style use case.
 ///
-/// The `Module` is thread-safe and safe to share across threads.
+/// Currently a `Module` does not implement any form of tiering or dynamic
+/// optimization of compiled code. Creation of a `Module` via [`Module::new`] or
+/// related APIs will perform the entire compilation step synchronously. When
+/// finished no further compilation will happen at runtime or later during
+/// execution of WebAssembly instances for example.
+///
+/// Compilation of WebAssembly by default goes through Cranelift and is
+/// recommended to be done once-per-module. The same WebAssembly binary need not
+/// be compiled multiple times and can instead used an embedder-cached result of
+/// the first call.
+///
+/// `Module` is thread-safe and safe to share across threads.
 ///
 /// ## Modules and `Clone`
 ///
@@ -98,6 +105,25 @@ pub use registry::{
 /// # }
 /// ```
 ///
+/// Serializing and deserializing a module looks like:
+///
+/// ```no_run
+/// # use wasmtime::*;
+/// # fn main() -> anyhow::Result<()> {
+/// let engine = Engine::default();
+/// # let wasm_bytes: Vec<u8> = Vec::new();
+/// let module = Module::new(&engine, &wasm_bytes)?;
+/// let module_bytes = module.serialize()?;
+///
+/// // ... can save `module_bytes` to disk or other storage ...
+///
+/// // recreate the module from the serialized bytes. For the `unsafe` bits
+/// // see the documentation of `deserialize`.
+/// let module = unsafe { Module::deserialize(&engine, &module_bytes)? };
+/// # Ok(())
+/// # }
+/// ```
+///
 /// [`Config`]: crate::Config
 #[derive(Clone)]
 pub struct Module {
@@ -124,7 +150,7 @@ struct ModuleInner {
     /// image this is a pretty expensive operation, so by deferring it this
     /// improves memory usage for modules that are created but may not ever be
     /// instantiated.
-    memory_images: OnceCell<Option<ModuleMemoryImages>>,
+    memory_images: OnceLock<Option<ModuleMemoryImages>>,
 
     /// Flag indicating whether this module can be serialized or not.
     serializable: bool,
@@ -133,14 +159,22 @@ struct ModuleInner {
     offsets: VMOffsets<HostPtr>,
 }
 
-impl std::fmt::Debug for Module {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+impl fmt::Debug for Module {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("Module")
             .field("name", &self.name())
             .finish_non_exhaustive()
     }
 }
 
+impl fmt::Debug for ModuleInner {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.debug_struct("ModuleInner")
+            .field("name", &self.module.module().name.as_ref())
+            .finish_non_exhaustive()
+    }
+}
+
 impl Module {
     /// Creates a new WebAssembly `Module` from the given in-memory `bytes`.
     ///
@@ -157,9 +191,8 @@ impl Module {
     /// loaded into memory all at once, this API does not support streaming
     /// compilation of a module.
     ///
-    /// If the module has not been already been compiled, the WebAssembly binary will
-    /// be decoded and validated. It will also be compiled according to the
-    /// configuration of the provided `engine`.
+    /// The WebAssembly binary will be decoded and validated. It will also be
+    /// compiled according to the configuration of the provided `engine`.
     ///
     /// # Errors
     ///
@@ -208,12 +241,10 @@ impl Module {
     /// # }
     /// ```
     #[cfg(any(feature = "cranelift", feature = "winch"))]
-    #[cfg_attr(nightlydoc, doc(cfg(any(feature = "cranelift", feature = "winch"))))]
     pub fn new(engine: &Engine, bytes: impl AsRef<[u8]>) -> Result<Module> {
-        let bytes = bytes.as_ref();
-        #[cfg(feature = "wat")]
-        let bytes = wat::parse_bytes(bytes)?;
-        Self::from_binary(engine, &bytes)
+        crate::CodeBuilder::new(engine)
+            .wasm(bytes.as_ref(), None)?
+            .compile_module()
     }
 
     /// Creates a new WebAssembly `Module` from the contents of the given
@@ -244,28 +275,11 @@ impl Module {
     /// # Ok(())
     /// # }
     /// ```
-    #[cfg(any(feature = "cranelift", feature = "winch"))]
-    #[cfg_attr(nightlydoc, doc(cfg(any(feature = "cranelift", feature = "winch"))))]
+    #[cfg(all(feature = "std", any(feature = "cranelift", feature = "winch")))]
     pub fn from_file(engine: &Engine, file: impl AsRef<Path>) -> Result<Module> {
-        let file = file.as_ref();
-        match Self::new(
-            engine,
-            &fs::read(file)
-                .with_context(|| format!("failed to read input file: {}", file.display()))?,
-        ) {
-            Ok(m) => Ok(m),
-            Err(e) => {
-                cfg_if::cfg_if! {
-                    if #[cfg(feature = "wat")] {
-                        let mut e = e.downcast::<wat::Error>()?;
-                        e.set_path(file);
-                        bail!(e)
-                    } else {
-                        Err(e)
-                    }
-                }
-            }
-        }
+        crate::CodeBuilder::new(engine)
+            .wasm_file(file.as_ref())?
+            .compile_module()
     }
 
     /// Creates a new WebAssembly `Module` from the given in-memory `binary`
@@ -300,56 +314,11 @@ impl Module {
     /// # }
     /// ```
     #[cfg(any(feature = "cranelift", feature = "winch"))]
-    #[cfg_attr(nightlydoc, doc(cfg(any(feature = "cranelift", feature = "winch"))))]
     pub fn from_binary(engine: &Engine, binary: &[u8]) -> Result<Module> {
-        use crate::{compile::build_artifacts, instantiate::MmapVecWrapper};
-
-        engine
-            .check_compatible_with_native_host()
-            .context("compilation settings are not compatible with the native host")?;
-
-        cfg_if::cfg_if! {
-            if #[cfg(feature = "cache")] {
-                let state = (HashedEngineCompileEnv(engine), binary);
-                let (code, info_and_types) = wasmtime_cache::ModuleCacheEntry::new(
-                    "wasmtime",
-                    engine.cache_config(),
-                )
-                .get_data_raw(
-                    &state,
-
-                    // Cache miss, compute the actual artifacts
-                    |(engine, wasm)| -> Result<_> {
-                        let (mmap, info) = build_artifacts::<MmapVecWrapper>(engine.0, wasm)?;
-                        let code = publish_mmap(mmap.0)?;
-                        Ok((code, info))
-                    },
-
-                    // Implementation of how to serialize artifacts
-                    |(_engine, _wasm), (code, _info_and_types)| {
-                        Some(code.mmap().to_vec())
-                    },
-
-                    // Cache hit, deserialize the provided artifacts
-                    |(engine, _wasm), serialized_bytes| {
-                        let code = engine.0.load_code_bytes(&serialized_bytes, ObjectKind::Module).ok()?;
-                        Some((code, None))
-                    },
-                )?;
-            } else {
-                let (mmap, info_and_types) = build_artifacts::<MmapVecWrapper>(engine, binary)?;
-                let code = publish_mmap(mmap.0)?;
-            }
-        };
-
-        let info_and_types = info_and_types.map(|(info, types)| (info, types.into()));
-        return Self::from_parts(engine, code, info_and_types);
-
-        fn publish_mmap(mmap: MmapVec) -> Result<Arc<CodeMemory>> {
-            let mut code = CodeMemory::new(mmap)?;
-            code.publish()?;
-            Ok(Arc::new(code))
-        }
+        crate::CodeBuilder::new(engine)
+            .wasm(binary, None)?
+            .wat(false)?
+            .compile_module()
     }
 
     /// Creates a new WebAssembly `Module` from the contents of the given `file`
@@ -373,8 +342,7 @@ impl Module {
     /// This is because the file is mapped into memory and lazily loaded pages
     /// reflect the current state of the file, not necessarily the original
     /// state of the file.
-    #[cfg(any(feature = "cranelift", feature = "winch"))]
-    #[cfg_attr(nightlydoc, doc(cfg(any(feature = "cranelift", feature = "winch"))))]
+    #[cfg(all(feature = "std", any(feature = "cranelift", feature = "winch")))]
     pub unsafe fn from_trusted_file(engine: &Engine, file: impl AsRef<Path>) -> Result<Module> {
         let mmap = MmapVec::from_file(file.as_ref())?;
         if &mmap[0..4] == b"\x7fELF" {
@@ -382,7 +350,9 @@ impl Module {
             return Module::from_parts(engine, code, None);
         }
 
-        Module::new(engine, &*mmap)
+        crate::CodeBuilder::new(engine)
+            .wasm(&mmap, Some(file.as_ref()))?
+            .compile_module()
     }
 
     /// Deserializes an in-memory compiled module previously created with
@@ -453,21 +423,22 @@ impl Module {
     /// entire lifetime of the [`Module`] returned. Any changes to the file on
     /// disk may change future instantiations of the module to be incorrect.
     /// This is because the file is mapped into memory and lazily loaded pages
-    /// reflect the current state of the file, not necessarily the origianl
+    /// reflect the current state of the file, not necessarily the original
     /// state of the file.
+    #[cfg(feature = "std")]
     pub unsafe fn deserialize_file(engine: &Engine, path: impl AsRef<Path>) -> Result<Module> {
         let code = engine.load_code_file(path.as_ref(), ObjectKind::Module)?;
         Module::from_parts(engine, code, None)
     }
 
     /// Entrypoint for creating a `Module` for all above functions, both
-    /// of the AOT and jit-compiled cateogries.
+    /// of the AOT and jit-compiled categories.
     ///
     /// In all cases the compilation artifact, `code_memory`, is provided here.
     /// The `info_and_types` argument is `None` when a module is being
     /// deserialized from a precompiled artifact or it's `Some` if it was just
     /// compiled and the values are already available.
-    fn from_parts(
+    pub(crate) fn from_parts(
         engine: &Engine,
         code_memory: Arc<CodeMemory>,
         info_and_types: Option<(CompiledModuleInfo, ModuleTypes)>,
@@ -477,7 +448,7 @@ impl Module {
         // already.
         let (info, types) = match info_and_types {
             Some((info, types)) => (info, types),
-            None => bincode::deserialize(code_memory.wasmtime_info())?,
+            None => postcard::from_bytes(code_memory.wasmtime_info()).err2anyhow()?,
         };
 
         // Register function type signatures into the engine for the lifetime
@@ -488,7 +459,7 @@ impl Module {
         // Note that the unsafety here should be ok since the `trampolines`
         // field should only point to valid trampoline function pointers
         // within the text section.
-        let signatures = TypeCollection::new_for_module(engine.signatures(), &types);
+        let signatures = TypeCollection::new_for_module(engine, &types);
 
         // Package up all our data into a `CodeObject` and delegate to the final
         // step of module compilation.
@@ -502,12 +473,8 @@ impl Module {
         info: CompiledModuleInfo,
         serializable: bool,
     ) -> Result<Self> {
-        let module = CompiledModule::from_artifacts(
-            code.code_memory().clone(),
-            info,
-            engine.profiler(),
-            engine.unique_id_allocator(),
-        )?;
+        let module =
+            CompiledModule::from_artifacts(code.code_memory().clone(), info, engine.profiler())?;
 
         // Validate the module can be used with the current instance allocator.
         let offsets = VMOffsets::new(HostPtr, module.module());
@@ -519,7 +486,7 @@ impl Module {
             inner: Arc::new(ModuleInner {
                 engine: engine.clone(),
                 code,
-                memory_images: OnceCell::new(),
+                memory_images: OnceLock::new(),
                 module,
                 serializable,
                 offsets,
@@ -551,8 +518,8 @@ impl Module {
 
         let mut functions = Vec::new();
         for payload in Parser::new(0).parse_all(binary) {
-            let payload = payload?;
-            if let ValidPayload::Func(a, b) = validator.payload(&payload)? {
+            let payload = payload.err2anyhow()?;
+            if let ValidPayload::Func(a, b) = validator.payload(&payload).err2anyhow()? {
                 functions.push((a, b));
             }
             if let wasmparser::Payload::Version { encoding, .. } = &payload {
@@ -562,13 +529,15 @@ impl Module {
             }
         }
 
-        engine.run_maybe_parallel(functions, |(validator, body)| {
-            // FIXME: it would be best here to use a rayon-specific parallel
-            // iterator that maintains state-per-thread to share the function
-            // validator allocations (`Default::default` here) across multiple
-            // functions.
-            validator.into_validator(Default::default()).validate(&body)
-        })?;
+        engine
+            .run_maybe_parallel(functions, |(validator, body)| {
+                // FIXME: it would be best here to use a rayon-specific parallel
+                // iterator that maintains state-per-thread to share the function
+                // validator allocations (`Default::default` here) across multiple
+                // functions.
+                validator.into_validator(Default::default()).validate(&body)
+            })
+            .err2anyhow()?;
         Ok(())
     }
 
@@ -582,7 +551,6 @@ impl Module {
     /// this method can be useful to get the serialized version without
     /// compiling twice.
     #[cfg(any(feature = "cranelift", feature = "winch"))]
-    #[cfg_attr(nightlydoc, doc(cfg(any(feature = "cranelift", feature = "winch"))))]
     pub fn serialize(&self) -> Result<Vec<u8>> {
         // The current representation of compiled modules within a compiled
         // component means that it cannot be serialized. The mmap returned here
@@ -611,11 +579,11 @@ impl Module {
         &self.inner.module
     }
 
-    fn code_object(&self) -> &Arc<CodeObject> {
+    pub(crate) fn code_object(&self) -> &Arc<CodeObject> {
         &self.inner.code
     }
 
-    pub(crate) fn env_module(&self) -> &wasmtime_environ::Module {
+    pub(crate) fn env_module(&self) -> &Arc<wasmtime_environ::Module> {
         self.compiled_module().module()
     }
 
@@ -707,9 +675,15 @@ impl Module {
     ) -> impl ExactSizeIterator<Item = ImportType<'module>> + 'module {
         let module = self.compiled_module().module();
         let types = self.types();
+        let engine = self.engine();
         module
             .imports()
-            .map(move |(module, field, ty)| ImportType::new(module, field, ty, types))
+            .map(move |(imp_mod, imp_field, mut ty)| {
+                ty.canonicalize_for_runtime_usage(&mut |i| {
+                    self.signatures().shared_type(i).unwrap()
+                });
+                ImportType::new(imp_mod, imp_field, ty, types, engine)
+            })
             .collect::<Vec<_>>()
             .into_iter()
     }
@@ -773,8 +747,9 @@ impl Module {
     ) -> impl ExactSizeIterator<Item = ExportType<'module>> + 'module {
         let module = self.compiled_module().module();
         let types = self.types();
+        let engine = self.engine();
         module.exports.iter().map(move |(name, entity_index)| {
-            ExportType::new(name, module.type_of(*entity_index), types)
+            ExportType::new(name, module.type_of(*entity_index), types, engine)
         })
     }
 
@@ -825,6 +800,7 @@ impl Module {
         let module = self.compiled_module().module();
         let entity_index = module.exports.get(name)?;
         Some(ExternType::from_wasmtime(
+            self.engine(),
             self.types(),
             &module.type_of(*entity_index),
         ))
@@ -923,15 +899,7 @@ impl Module {
         }
     }
 
-    /// Returns the `ModuleInner` cast as `ModuleRuntimeInfo` for use
-    /// by the runtime.
-    pub(crate) fn runtime_info(&self) -> Arc<dyn wasmtime_runtime::ModuleRuntimeInfo> {
-        // N.B.: this needs to return a clone because we cannot
-        // statically cast the &Arc<ModuleInner> to &Arc<dyn Trait...>.
-        self.inner.clone()
-    }
-
-    pub(crate) fn module_info(&self) -> &dyn wasmtime_runtime::ModuleInfo {
+    pub(crate) fn module_info(&self) -> &dyn crate::runtime::vm::ModuleInfo {
         &*self.inner
     }
 
@@ -949,8 +917,8 @@ impl Module {
     ///
     /// It is not safe to modify the memory in this range, nor is it safe to
     /// modify the protections of memory in this range.
-    pub fn image_range(&self) -> Range<usize> {
-        self.compiled_module().image_range()
+    pub fn image_range(&self) -> Range<*const u8> {
+        self.compiled_module().mmap().image_range()
     }
 
     /// Force initialization of copy-on-write images to happen here-and-now
@@ -971,7 +939,7 @@ impl Module {
     /// as a performance optimization if required but is otherwise handled
     /// automatically.
     pub fn initialize_copy_on_write_image(&self) -> Result<()> {
-        self.inner.memory_images()?;
+        self.memory_images()?;
         Ok(())
     }
 
@@ -1032,31 +1000,87 @@ impl Module {
         self.code_object().code_memory().text()
     }
 
-    /// Get the locations of functions in this module's `.text` section.
+    /// Get information about functions in this module's `.text` section: their
+    /// index, name, and offset+length.
     ///
-    /// Each function's location is a (`.text` section offset, length) pair.
-    pub fn function_locations<'a>(&'a self) -> impl ExactSizeIterator<Item = (usize, usize)> + 'a {
-        self.compiled_module().finished_functions().map(|(f, _)| {
-            let loc = self.compiled_module().func_loc(f);
-            (loc.start as usize, loc.length as usize)
+    /// Results are yielded in a ModuleFunction struct.
+    pub fn functions<'a>(&'a self) -> impl ExactSizeIterator<Item = ModuleFunction> + 'a {
+        let module = self.compiled_module();
+        module.finished_functions().map(|(idx, _)| {
+            let loc = module.func_loc(idx);
+            let idx = module.module().func_index(idx);
+            ModuleFunction {
+                index: idx,
+                name: module.func_name(idx).map(|n| n.to_string()),
+                offset: loc.start as usize,
+                len: loc.length as usize,
+            }
         })
     }
 
     pub(crate) fn id(&self) -> CompiledModuleId {
         self.inner.module.unique_id()
     }
-}
 
-impl ModuleInner {
-    fn memory_images(&self) -> Result<Option<&ModuleMemoryImages>> {
+    pub(crate) fn offsets(&self) -> &VMOffsets<HostPtr> {
+        &self.inner.offsets
+    }
+
+    /// Return the address, in memory, of the trampoline that allows Wasm to
+    /// call a array function of the given signature.
+    pub(crate) fn wasm_to_array_trampoline(
+        &self,
+        signature: VMSharedTypeIndex,
+    ) -> Option<NonNull<VMWasmCallFunction>> {
+        log::trace!("Looking up trampoline for {signature:?}");
+        let trampoline_shared_ty = self.inner.engine.signatures().trampoline_type(signature);
+        let trampoline_module_ty = self
+            .inner
+            .code
+            .signatures()
+            .trampoline_type(trampoline_shared_ty)?;
+        debug_assert!(self
+            .inner
+            .engine
+            .signatures()
+            .borrow(
+                self.inner
+                    .code
+                    .signatures()
+                    .shared_type(trampoline_module_ty)
+                    .unwrap()
+            )
+            .unwrap()
+            .unwrap_func()
+            .is_trampoline_type());
+
+        let ptr = self
+            .compiled_module()
+            .wasm_to_array_trampoline(trampoline_module_ty)
+            .as_ptr()
+            .cast::<VMWasmCallFunction>()
+            .cast_mut();
+        Some(NonNull::new(ptr).unwrap())
+    }
+
+    pub(crate) fn memory_images(&self) -> Result<Option<&ModuleMemoryImages>> {
         let images = self
+            .inner
             .memory_images
-            .get_or_try_init(|| memory_images(&self.engine, &self.module))?
+            .get_or_try_init(|| memory_images(&self.inner.engine, &self.inner.module))?
             .as_ref();
         Ok(images)
     }
 }
 
+/// Describes a function for a given module.
+pub struct ModuleFunction {
+    pub index: wasmtime_environ::FuncIndex,
+    pub name: Option<String>,
+    pub offset: usize,
+    pub len: usize,
+}
+
 impl Drop for ModuleInner {
     fn drop(&mut self) {
         // When a `Module` is being dropped that means that it's no longer
@@ -1085,105 +1109,7 @@ fn _assert_send_sync() {
     _assert::<Module>();
 }
 
-/// This is a helper struct used when caching to hash the state of an `Engine`
-/// used for module compilation.
-///
-/// The hash computed for this structure is used to key the global wasmtime
-/// cache and dictates whether artifacts are reused. Consequently the contents
-/// of this hash dictate when artifacts are or aren't re-used.
-#[cfg(any(feature = "cranelift", feature = "winch"))]
-pub(crate) struct HashedEngineCompileEnv<'a>(pub &'a Engine);
-
-#[cfg(any(feature = "cranelift", feature = "winch"))]
-impl std::hash::Hash for HashedEngineCompileEnv<'_> {
-    fn hash<H: std::hash::Hasher>(&self, hasher: &mut H) {
-        // Hash the compiler's state based on its target and configuration.
-        let compiler = self.0.compiler();
-        compiler.triple().hash(hasher);
-        compiler.flags().hash(hasher);
-        compiler.isa_flags().hash(hasher);
-
-        // Hash configuration state read for compilation
-        let config = self.0.config();
-        self.0.tunables().hash(hasher);
-        config.features.hash(hasher);
-        config.wmemcheck.hash(hasher);
-
-        // Catch accidental bugs of reusing across crate versions.
-        config.module_version.hash(hasher);
-    }
-}
-
-impl wasmtime_runtime::ModuleRuntimeInfo for ModuleInner {
-    fn module(&self) -> &Arc<wasmtime_environ::Module> {
-        self.module.module()
-    }
-
-    fn function(&self, index: DefinedFuncIndex) -> NonNull<VMWasmCallFunction> {
-        let ptr = self
-            .module
-            .finished_function(index)
-            .as_ptr()
-            .cast::<VMWasmCallFunction>()
-            .cast_mut();
-        NonNull::new(ptr).unwrap()
-    }
-
-    fn native_to_wasm_trampoline(
-        &self,
-        index: DefinedFuncIndex,
-    ) -> Option<NonNull<VMNativeCallFunction>> {
-        let ptr = self
-            .module
-            .native_to_wasm_trampoline(index)?
-            .as_ptr()
-            .cast::<VMNativeCallFunction>()
-            .cast_mut();
-        Some(NonNull::new(ptr).unwrap())
-    }
-
-    fn array_to_wasm_trampoline(&self, index: DefinedFuncIndex) -> Option<VMArrayCallFunction> {
-        let ptr = self.module.array_to_wasm_trampoline(index)?.as_ptr();
-        Some(unsafe { mem::transmute::<*const u8, VMArrayCallFunction>(ptr) })
-    }
-
-    fn wasm_to_native_trampoline(
-        &self,
-        signature: VMSharedTypeIndex,
-    ) -> Option<NonNull<VMWasmCallFunction>> {
-        let sig = self.code.signatures().module_local_type(signature)?;
-        let ptr = self
-            .module
-            .wasm_to_native_trampoline(sig)
-            .as_ptr()
-            .cast::<VMWasmCallFunction>()
-            .cast_mut();
-        Some(NonNull::new(ptr).unwrap())
-    }
-
-    fn memory_image(&self, memory: DefinedMemoryIndex) -> Result<Option<&Arc<MemoryImage>>> {
-        let images = self.memory_images()?;
-        Ok(images.and_then(|images| images.get_memory_image(memory)))
-    }
-
-    fn unique_id(&self) -> Option<CompiledModuleId> {
-        Some(self.module.unique_id())
-    }
-
-    fn wasm_data(&self) -> &[u8] {
-        self.module.code_memory().wasm_data()
-    }
-
-    fn type_ids(&self) -> &[VMSharedTypeIndex] {
-        self.code.signatures().as_module_map().values().as_slice()
-    }
-
-    fn offsets(&self) -> &VMOffsets<HostPtr> {
-        &self.offsets
-    }
-}
-
-impl wasmtime_runtime::ModuleInfo for ModuleInner {
+impl crate::runtime::vm::ModuleInfo for ModuleInner {
     fn lookup_stack_map(&self, pc: usize) -> Option<&wasmtime_environ::StackMap> {
         let text_offset = pc - self.module.text().as_ptr() as usize;
         let (index, func_offset) = self.module.func_by_text_offset(text_offset)?;
@@ -1209,88 +1135,6 @@ impl wasmtime_runtime::ModuleInfo for ModuleInner {
     }
 }
 
-/// A barebones implementation of ModuleRuntimeInfo that is useful for
-/// cases where a purpose-built environ::Module is used and a full
-/// CompiledModule does not exist (for example, for tests or for the
-/// default-callee instance).
-pub(crate) struct BareModuleInfo {
-    module: Arc<wasmtime_environ::Module>,
-    one_signature: Option<VMSharedTypeIndex>,
-    offsets: VMOffsets<HostPtr>,
-}
-
-impl BareModuleInfo {
-    pub(crate) fn empty(module: Arc<wasmtime_environ::Module>) -> Self {
-        BareModuleInfo::maybe_imported_func(module, None)
-    }
-
-    pub(crate) fn maybe_imported_func(
-        module: Arc<wasmtime_environ::Module>,
-        one_signature: Option<VMSharedTypeIndex>,
-    ) -> Self {
-        BareModuleInfo {
-            offsets: VMOffsets::new(HostPtr, &module),
-            module,
-            one_signature,
-        }
-    }
-
-    pub(crate) fn into_traitobj(self) -> Arc<dyn wasmtime_runtime::ModuleRuntimeInfo> {
-        Arc::new(self)
-    }
-}
-
-impl wasmtime_runtime::ModuleRuntimeInfo for BareModuleInfo {
-    fn module(&self) -> &Arc<wasmtime_environ::Module> {
-        &self.module
-    }
-
-    fn function(&self, _index: DefinedFuncIndex) -> NonNull<VMWasmCallFunction> {
-        unreachable!()
-    }
-
-    fn array_to_wasm_trampoline(&self, _index: DefinedFuncIndex) -> Option<VMArrayCallFunction> {
-        unreachable!()
-    }
-
-    fn native_to_wasm_trampoline(
-        &self,
-        _index: DefinedFuncIndex,
-    ) -> Option<NonNull<VMNativeCallFunction>> {
-        unreachable!()
-    }
-
-    fn wasm_to_native_trampoline(
-        &self,
-        _signature: VMSharedTypeIndex,
-    ) -> Option<NonNull<VMWasmCallFunction>> {
-        unreachable!()
-    }
-
-    fn memory_image(&self, _memory: DefinedMemoryIndex) -> Result<Option<&Arc<MemoryImage>>> {
-        Ok(None)
-    }
-
-    fn unique_id(&self) -> Option<CompiledModuleId> {
-        None
-    }
-
-    fn wasm_data(&self) -> &[u8] {
-        &[]
-    }
-
-    fn type_ids(&self) -> &[VMSharedTypeIndex] {
-        match &self.one_signature {
-            Some(id) => std::slice::from_ref(id),
-            None => &[],
-        }
-    }
-
-    fn offsets(&self) -> &VMOffsets<HostPtr> {
-        &self.offsets
-    }
-}
-
 /// Helper method to construct a `ModuleMemoryImages` for an associated
 /// `CompiledModule`.
 fn memory_images(engine: &Engine, module: &CompiledModule) -> Result<Option<ModuleMemoryImages>> {
